pragma solidity >=0.5.5;

/* Simple one shot, time locked and conditional on two-party release escrow smart contract */

contract Escrow2 {

    enum State {AwaitingDeposit, DepositPlaced, Withdrawn}

    /* Recall:
     * net(addr) := (funds sent from addr to this) - (funds sent from this to addr)
     */
    /*@ invariant
      @   sender != receiver,
      @   amountInEscrow == net(sender) + net(receiver),
      @   // needed for postcond of placeInEscrow:
      @   state == State.AwaitingDeposit ==> net(sender) == 0,
      @   state != State.DepositPlaced ==> amountInEscrow == 0,
      @   // Probably valid (may be unimportat in this example?):
      @   (\forall address a;
      @                (a != sender && a != receiver && a != address(this))
      @            ==> net(a) == 0);
      @*/

    // why not /* payable */ sender as in v1?
    address public sender;
    address /* payable */ public receiver;

    uint public delayUntilRelease;
    uint public releaseTime;

    uint public amountInEscrow;
    bool public releasedBySender;
    bool public releasedByReceiver;

    State public state;

    modifier by(address _address) {
        require(msg.sender == _address);
        _;
    }

    modifier stateIs(State _state) {
        require(state == _state);
        _;
    }

    constructor (address _sender, address /* payable */ _receiver, uint _delayUntilRelease) public {
        // Set parameters of escrow contract
        // forgotten  at first:
        require(sender != receiver);
        sender = _sender;
        receiver = _receiver;
        delayUntilRelease = _delayUntilRelease;

        releasedBySender = false;
        releasedByReceiver = false;

        // Set contract state
        state = State.AwaitingDeposit;
    }

    /*@ succeeds_only_if
      @   msg.sender == sender,           // by(sender)
      @   state == State.AwaitingDeposit, // stateIs(State.AwaitingDeposit)
      @   msg.value > 0;
      @ after_success
      @   net(sender) == msg.value,
      @   state == State.DepositPlaced;
      @*/
    function placeInEscrow() public by(sender) stateIs(State.AwaitingDeposit) payable {
        require (msg.value > 0);

        // Update parameters of escrow contract
        amountInEscrow = msg.value;
        releaseTime = now + delayUntilRelease;

        // Set contract state
        state = State.DepositPlaced;
    }

    /*@ succeeds_only_if
      @   state == State.DepositPlaced, // stateIs(State.DepositPlaced)
      @ after_success
      @   msg.sender == sender ==> releasedBySender,
      @   msg.sender == receiver ==> releasedByReceiver,
      @*/
    function releaseEscrow() public stateIs(State.DepositPlaced) {
        if (msg.sender == sender)   { releasedBySender   = true; }
        if (msg.sender == receiver) { releasedByReceiver = true; }
    }

    /*@ succeeds_only_if
      @   releasedByReceiver,
      @   releasedBySender,
      @   msg.sender == receiver,       // by(receiver)
      @   state == State.DepositPlaced, // stateIs(State.DepositPlaced)
      @ after_success
      @   net(receiver) == -net(sender),
      @   state == State.Withdrawn;
      @*/
    // invariant broken at transfer
    function withdrawFromEscrow() public by(receiver) stateIs(State.DepositPlaced) {
        require (now >= releaseTime);
        require (releasedByReceiver && releasedBySender);

        // Set contract state
        state = State.Withdrawn;

        // Send money
        receiver.transfer(amountInEscrow);

        // Set internal parameters of smart contract
        amountInEscrow = 0;

    }
}
