pragma solidity >=0.5.5;

contract PiggyBank1 {
    enum PiggyBankState { Unused, InUse, Broken }

    /* Recall:
     * net(addr) := (funds sent from addr to this) - (funds sent from this to addr)
     */
    /*@ invariant
      @   state != PiggyBankState.Broken ==> balance == net(owner),
      @   state == PiggyBankState.Broken ==> net(owner) == 0;
      @*/

    address /* payable */ public owner;
    PiggyBankState public state;
    uint public timeOfFirstDeposit;
    uint public balance;

    /*@ after_success
      @   state == PiggyBankState.Unused,
      @   balance == 0,
      @   owner == _owner;
      @*/
    constructor (address /* payable */ _owner) public {
        owner = _owner;
        state = PiggyBankState.Unused;
        balance = 0;
    }

    modifier byOwner {
        require (msg.sender == owner);
        _;
    }

    modifier notBroken {
        require (state != PiggyBankState.Broken);
        _;
    }

    modifier inUse {
        require (state == PiggyBankState.InUse);
        _;
    }

    /*@ succeeds_only_if
      @   msg.sender == owner,            // byOwner
      @   state != PiggyBankState.Broken; // notBroken
      @ after_success
      @   balance == net(owner),
      @   state == PiggyBankState.InUse;
      @*/
    function addMoney() byOwner notBroken public payable {
        balance = balance + msg.value;
        if (state == PiggyBankState.Unused) {
            state = PiggyBankState.InUse;
            timeOfFirstDeposit = now;
        }
    }

    /*@ succeeds_only_if
      @   msg.sender == owner,           // byOwner
      @   state == PiggyBankState.InUse; // inUse
      @ after_success
      @   net(owner) == 0,
      @   state == PiggyBankState.Broken;
      @*/
    function breakPiggyBank() byOwner inUse public {
        require (now >= timeOfFirstDeposit + 365 days);

        state = PiggyBankState.Broken;
        owner.transfer(balance);
    }

}
